@@ -65,16 +65,16 @@ static void DPKG_ATTR_NORET
 usage(const struct cmdinfo *cip, const char *value)
 {
   printf(_(
-"Usage: %s [<option> ...] <command>\n"
+"Usage: %s [<option>...] <command>\n"
 "\n"), BACKEND);
 
   printf(_(
 "Commands:\n"
 "  -b|--build <directory> [<deb>]   Build an archive.\n"
 "  -c|--contents <deb>              List contents.\n"
-"  -I|--info <deb> [<cfile> ...]    Show info to stdout.\n"
+"  -I|--info <deb> [<cfile>...]     Show info to stdout.\n"
 "  -W|--show <deb>                  Show information on package(s)\n"
-"  -f|--field <deb> [<cfield> ...]  Show field(s) to stdout.\n"
+"  -f|--field <deb> [<cfield>...]   Show field(s) to stdout.\n"
 "  -e|--control <deb> [<directory>] Extract control info.\n"
 "  -x|--extract <deb> <directory>   Extract files.\n"
 "  -X|--vextract <deb> <directory>  Extract & list files.\n"
@@ -102,14 +102,13 @@ usage(const struct cmdinfo *cip, const char *value)
 "      --showformat=<format>        Use alternative format for --show.\n"
 "      --deb-format=<format>        Select archive format.\n"
 "                                     Allowed values: 0.939000, 2.0 (default).\n"
-"      --old                        Legacy alias for '--deb-format=0.939000'.\n"
-"      --new                        Legacy alias for '--deb-format=2.0'.\n"
 "      --nocheck                    Suppress control file check (build bad\n"
 "                                     packages).\n"
-"      --uniform-compression        Use the compression params on all members.\n"
+"      --root-owner-group           Forces the owner and groups to root.\n"
+"      --[no-]uniform-compression   Use the compression params on all members.\n"
 "  -z#                              Set the compression level when building.\n"
 "  -Z<type>                         Set the compression type used when building.\n"
-"                                     Allowed types: gzip, xz, bzip2, none.\n"
+"                                     Allowed types: gzip, xz, zstd, none.\n"
 "  -S<strategy>                     Set the compression strategy when building.\n"
 "                                     Allowed values: none; extreme (xz);\n"
 "                                     filtered, huffman, rle, fixed (gzip).\n"
@@ -142,7 +141,8 @@ static const char printforhelp[] =
 int debugflag = 0;
 int nocheckflag = 0;
 int opt_verbose = 0;
-int opt_uniform_compression = 0;
+int opt_root_owner_group = 0;
+int opt_uniform_compression = 1;
 
 struct deb_version deb_format = DEB_VERSION(2, 0);
 
@@ -162,24 +162,6 @@ set_deb_format(const struct cmdinfo *cip, const char *value)
     badusage(_("unknown deb format version: %s"), value);
 }
 
-static void
-set_deb_old(const struct cmdinfo *cip, const char *value)
-{
-  deb_format = DEB_VERSION_OBJECT(0, 939000);
-
-  warning(_("obsolete option '--%s'; please use '--%s' instead"),
-          cip->olong, "deb-format=0.939000");
-}
-
-static void
-set_deb_new(const struct cmdinfo *cip, const char *value)
-{
-  deb_format = DEB_VERSION_OBJECT(2, 0);
-
-  warning(_("obsolete option '--%s'; please use '--%s' instead"),
-          cip->olong, "deb-format=2.0");
-}
-
 struct compress_params compress_params = {
   .type = DPKG_DEB_DEFAULT_COMPRESSOR,
   .strategy = COMPRESSOR_STRATEGY_NONE,
@@ -210,16 +192,12 @@ static void
 set_compress_type(const struct cmdinfo *cip, const char *value)
 {
   compress_params.type = compressor_find_by_name(value);
-  if (compress_params.type == COMPRESSOR_TYPE_ZSTD) {
-    compress_params.type = COMPRESSOR_TYPE_UNKNOWN;
-    badusage(_("only decompression is supported for '%s'!"), value);
-  }
   if (compress_params.type == COMPRESSOR_TYPE_UNKNOWN)
     badusage(_("unknown compression type '%s'!"), value);
   if (compress_params.type == COMPRESSOR_TYPE_LZMA)
-    warning(_("deprecated compression type '%s'; use xz instead"), value);
+    badusage(_("obsolete compression type '%s'; use xz instead"), value);
   if (compress_params.type == COMPRESSOR_TYPE_BZIP2)
-    warning(_("deprecated compression type '%s'; use xz or gzip instead"), value);
+    badusage(_("obsolete compression type '%s'; use xz or gzip instead"), value);
 }
 
 static const struct cmdinfo cmdinfos[]= {
@@ -236,12 +214,12 @@ static const struct cmdinfo cmdinfos[]= {
   ACTION("show",          'W', 0, do_showinfo),
 
   { "deb-format",    0,   1, NULL,           NULL,         set_deb_format   },
-  { "new",           0,   0, NULL,           NULL,         set_deb_new      },
-  { "old",           0,   0, NULL,           NULL,         set_deb_old      },
   { "debug",         'D', 0, &debugflag,     NULL,         NULL,          1 },
   { "verbose",       'v', 0, &opt_verbose,   NULL,         NULL,          1 },
   { "nocheck",       0,   0, &nocheckflag,   NULL,         NULL,          1 },
+  { "root-owner-group",    0, 0, &opt_root_owner_group,    NULL, NULL,    1 },
   { "uniform-compression", 0, 0, &opt_uniform_compression, NULL, NULL,    1 },
+  { "no-uniform-compression", 0, 0, &opt_uniform_compression, NULL, NULL, 0 },
   { NULL,            'z', 1, NULL,           NULL,         set_compress_level },
   { NULL,            'Z', 1, NULL,           NULL,         set_compress_type  },
   { NULL,            'S', 1, NULL,           NULL,         set_compress_strategy },
@@ -267,6 +245,7 @@ int main(int argc, const char *const *argv) {
   if (opt_uniform_compression &&
       (compress_params.type != COMPRESSOR_TYPE_NONE &&
        compress_params.type != COMPRESSOR_TYPE_GZIP &&
+       compress_params.type != COMPRESSOR_TYPE_ZSTD &&
        compress_params.type != COMPRESSOR_TYPE_XZ))
     badusage(_("unsupported compression type '%s' with uniform compression"),
              compressor_get_name(compress_params.type));
@@ -274,6 +253,7 @@ int main(int argc, const char *const *argv) {
   ret = cipaction->action(argv);
 
   dpkg_program_done();
+  dpkg_locales_done();
 
   return ret;
 }
