@@ -2,7 +2,7 @@
  * dselect - Debian package maintenance user interface
  * basecmds.cc - base list keyboard commands display
  *
- * Copyright © 1994,1995 Ian Jackson <ian@chiark.greenend.org.uk>
+ * Copyright © 1994,1995 Ian Jackson <ijackson@chiark.greenend.org.uk>
  * Copyright © 2000,2001 Wichert Akkerman <wakkerma@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
@@ -76,13 +76,6 @@ void baselist::kd_bottom() {
 }
 
 void baselist::kd_redraw() {
-//#define RFSH(x) werase(x); redrawwin(x)
-//  RFSH(listpad);
-//  RFSH(infopad);
-//  RFSH(colheadspad);
-//  RFSH(thisstatepad);
-//  RFSH(titlewin);
-//  RFSH(whatinfowin); /* FIXME: why does ncurses need this? */
   clearok(curscr,TRUE);
   redrawall();
   debug(dbg_general, "baselist[%p]::kd_redraw() done", this);
@@ -107,7 +100,7 @@ baselist::matchsearch(int index)
 
   name = itemname(index);
   if (!name)
-    return false;	/* Skip things without a name (seperators). */
+    return false;	/* Skip things without a name (separators). */
 
   searchlen=strlen(searchstring);
   lendiff = strlen(name) - searchlen;
@@ -123,11 +116,11 @@ void baselist::kd_search() {
   strcpy(newsearchstring,searchstring);
   werase(querywin);
   mvwaddstr(querywin,0,0, _("Search for ? "));
-  echo(); /* FIXME: ncurses documentation or implementation. */
+  echo();
   if (wgetnstr(querywin,newsearchstring,sizeof(newsearchstring)-1) == ERR)
     searchstring[0]= 0;
-  raise(SIGWINCH); /* FIXME: ncurses and xterm arrow keys. */
-  noecho(); /* FIXME: ncurses. */
+  raise(SIGWINCH);
+  noecho();
   if (whatinfo_height) { touchwin(whatinfowin); refreshinfo(); }
   else if (info_height) { touchwin(infopad); refreshinfo(); }
   else if (thisstate_height) redrawthisstate();
@@ -198,8 +191,8 @@ _("Press ? for help menu, . for next topic, <space> to exit help."));
         mvaddstr(i+3,6, gettext(hme->msg->title));
       }
       mvaddstr(i+4,1,
-               _("Press a key from the list above, <space> or `q' to exit help,\n"
-               "  or `.' (full stop) to read each help page in turn. "));
+               _("Press a key from the list above, <space> or 'q' to exit help,\n"
+                 "  or '.' (full stop) to read each help page in turn. "));
       nextkey= helpmenu[0].key;
     }
     refresh();
