@@ -2,7 +2,7 @@
  * libdpkg - Debian packaging suite library routines
  * options.c - option parsing functions
  *
- * Copyright © 1994,1995 Ian Jackson <ian@chiark.greenend.org.uk>
+ * Copyright © 1994,1995 Ian Jackson <ijackson@chiark.greenend.org.uk>
  * Copyright © 2000,2002 Wichert Akkerman <wichert@deephackmode.org>
  * Copyright © 2008-2015 Guillem Jover <guillem@debian.org>
  *
@@ -41,11 +41,11 @@ static const char *printforhelp;
 void
 badusage(const char *fmt, ...)
 {
-  char buf[1024];
+  char *buf = NULL;
   va_list args;
 
   va_start(args, fmt);
-  vsnprintf(buf, sizeof(buf), fmt, args);
+  m_vasprintf(&buf, fmt, args);
   va_end(args);
 
   ohshit("%s\n\n%s", buf, gettext(printforhelp));
@@ -54,11 +54,11 @@ badusage(const char *fmt, ...)
 static void DPKG_ATTR_NORET DPKG_ATTR_PRINTF(3)
 config_error(const char *file_name, int line_num, const char *fmt, ...)
 {
-  char buf[1024];
+  char *buf = NULL;
   va_list args;
 
   va_start(args, fmt);
-  vsnprintf(buf, sizeof(buf), fmt, args);
+  m_vasprintf(&buf, fmt, args);
   va_end(args);
 
   ohshit(_("configuration error: %s:%d: %s"), file_name, line_num, buf);
@@ -83,14 +83,10 @@ dpkg_options_load_file(const char *fn, const struct cmdinfo *cmdinfos)
   while (fgets(linebuf, sizeof(linebuf), file)) {
     char *opt;
     const struct cmdinfo *cip;
-    int l;
 
     line_num++;
 
-    l = strlen(linebuf);
-    while (l && c_isspace(linebuf[l - 1]))
-      l--;
-    linebuf[l] = '\0';
+    str_rtrim_spaces(linebuf, linebuf + strlen(linebuf));
 
     if ((linebuf[0] == '#') || (linebuf[0] == '\0'))
       continue;
@@ -109,6 +105,8 @@ dpkg_options_load_file(const char *fn, const struct cmdinfo *cmdinfos)
     }
 
     for (cip=cmdinfos; cip->olong || cip->oshort; cip++) {
+      int l;
+
       if (!cip->olong) continue;
       if (strcmp(cip->olong, linebuf) == 0)
         break;
@@ -168,7 +166,7 @@ dpkg_options_load_dir(const char *prog, const struct cmdinfo *cmdinfos)
   struct dirent **dlist;
   int dlist_n, i;
 
-  m_asprintf(&dirname, "%s/%s.cfg.d", CONFIGDIR, prog);
+  dirname = str_fmt("%s/%s.cfg.d", CONFIGDIR, prog);
 
   dlist_n = scandir(dirname, &dlist, valid_config_filename, alphasort);
   if (dlist_n < 0) {
@@ -182,7 +180,7 @@ dpkg_options_load_dir(const char *prog, const struct cmdinfo *cmdinfos)
   for (i = 0; i < dlist_n; i++) {
     char *filename;
 
-    m_asprintf(&filename, "%s/%s", dirname, dlist[i]->d_name);
+    filename = str_fmt("%s/%s", dirname, dlist[i]->d_name);
     dpkg_options_load_file(filename, cmdinfos);
 
     free(dlist[i]);
@@ -200,13 +198,13 @@ dpkg_options_load(const char *prog, const struct cmdinfo *cmdinfos)
 
   dpkg_options_load_dir(prog, cmdinfos);
 
-  m_asprintf(&file, "%s/%s.cfg", CONFIGDIR, prog);
+  file = str_fmt("%s/%s.cfg", CONFIGDIR, prog);
   dpkg_options_load_file(file, cmdinfos);
   free(file);
 
   home = getenv("HOME");
   if (home != NULL) {
-    m_asprintf(&file, "%s/.%s.cfg", home, prog);
+    file = str_fmt("%s/.%s.cfg", home, prog);
     dpkg_options_load_file(file, cmdinfos);
     free(file);
   }
@@ -287,8 +285,12 @@ dpkg_options_parse_arg_int(const struct cmdinfo *cmd, const char *str)
 
   errno = 0;
   value = strtol(str, &end, 0);
-  if (str == end || *end || value < 0 || value > INT_MAX || errno != 0)
-    badusage(_("invalid integer for --%s: '%.250s'"), cmd->olong, str);
+  if (str == end || *end || value < 0 || value > INT_MAX || errno != 0) {
+    if (cmd->olong)
+      badusage(_("invalid integer for --%s: '%.250s'"), cmd->olong, str);
+    else
+      badusage(_("invalid integer for -%c: '%.250s'"), cmd->oshort, str);
+  }
 
   return value;
 }
@@ -311,9 +313,11 @@ option_short(int c)
 void
 setaction(const struct cmdinfo *cip, const char *value)
 {
-  if (cipaction)
+  if (cipaction && cip)
     badusage(_("conflicting actions -%c (--%s) and -%c (--%s)"),
              option_short(cip->oshort), cip->olong,
              option_short(cipaction->oshort), cipaction->olong);
   cipaction = cip;
+  if (cip->takesvalue == 2 && cip->sassignto)
+    *cipaction->sassignto = value;
 }
