@@ -3,7 +3,7 @@
  * tarfn.c - tar archive extraction functions
  *
  * Copyright © 1995 Bruce Perens
- * Copyright © 2007-2011, 2013-2015 Guillem Jover <guillem@debian.org>
+ * Copyright © 2007-2011, 2013-2017 Guillem Jover <guillem@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,6 +22,9 @@
 #include <config.h>
 #include <compat.h>
 
+#if HAVE_SYS_SYSMACROS_H
+#include <sys/sysmacros.h>
+#endif
 #include <sys/stat.h>
 
 #include <errno.h>
@@ -35,11 +38,29 @@
 
 #include <dpkg/macros.h>
 #include <dpkg/dpkg.h>
+#include <dpkg/i18n.h>
+#include <dpkg/error.h>
 #include <dpkg/tarfn.h>
 
 #define TAR_MAGIC_USTAR "ustar\0" "00"
 #define TAR_MAGIC_GNU   "ustar "  " \0"
 
+#define TAR_TYPE_SIGNED(t)	(!((t)0 < (t)-1))
+
+#define TAR_TYPE_MIN(t) \
+	(TAR_TYPE_SIGNED(t) ? \
+	 ~(t)TAR_TYPE_MAX(t) : \
+	 (t)0)
+#define TAR_TYPE_MAX(t) \
+	(TAR_TYPE_SIGNED(t) ? \
+	 ((((t)1 << (sizeof(t) * 8 - 2)) - 1) * 2 + 1) : \
+	 ~(t)0)
+
+#define TAR_ATOUL(str, type) \
+	(type)tar_atoul(str, sizeof(str), TAR_TYPE_MAX(type))
+#define TAR_ATOSL(str, type) \
+	(type)tar_atosl(str, sizeof(str), TAR_TYPE_MIN(type), TAR_TYPE_MAX(type))
+
 struct tar_header {
 	char name[100];
 	char mode[8];
@@ -62,34 +83,136 @@ struct tar_header {
 	char prefix[155];
 };
 
+static inline uintmax_t
+tar_ret_errno(int err, uintmax_t ret)
+{
+	errno = err;
+	return ret;
+}
+
 /**
  * Convert an ASCII octal string to an intmax_t.
  */
-static intmax_t
-OtoM(const char *s, int size)
+static uintmax_t
+tar_atol8(const char *s, size_t size)
 {
-	intmax_t n = 0;
+	const char *end = s + size;
+	uintmax_t n = 0;
 
-	while (*s == ' ') {
+	/* Old implementations might precede the value with spaces. */
+	while (s < end && *s == ' ')
 		s++;
+
+	if (s == end)
+		return tar_ret_errno(EINVAL, 0);
+
+	while (s < end) {
+		if (*s == '\0' || *s == ' ')
+			break;
+		if (*s < '0' || *s > '7')
+			return tar_ret_errno(ERANGE, 0);
+		n = (n * 010) + (*s++ - '0');
+	}
+
+	while (s < end) {
+		if (*s != '\0' && *s != ' ')
+			return tar_ret_errno(EINVAL, 0);
+		s++;
+	}
+
+	if (s < end)
+		return tar_ret_errno(EINVAL, 0);
+
+	return tar_ret_errno(0, n);
+}
+
+/**
+ * Convert a base-256 two-complement number to an intmax_t.
+ */
+static uintmax_t
+tar_atol256(const char *s, size_t size, intmax_t min, uintmax_t max)
+{
+	uintmax_t n = 0;
+	unsigned char c;
+	int sign;
+
+	/* The encoding always sets the first bit to one, so that it can be
+	 * distinguished from the ASCII encoding. For positive numbers we
+	 * need to reset it. For negative numbers we initialize n to -1. */
+	c = *s++;
+	if (c == 0x80)
+		c = 0;
+	else
+		n = ~(uintmax_t)0;
+	sign = c;
+
+	/* Check for overflows. */
+	while (size > sizeof(uintmax_t)) {
+		if (c != sign)
+			return tar_ret_errno(ERANGE, sign ? (uintmax_t)min : max);
+		c = *s++;
 		size--;
 	}
 
-	while (--size >= 0 && *s >= '0' && *s <= '7')
-		n = (n * 010) + (*s++ - '0');
+	if ((c & 0x80) != (sign & 0x80))
+		return tar_ret_errno(ERANGE, sign ? (uintmax_t)min : max);
+
+	for (;;) {
+		n = (n << 8) | c;
+		if (--size == 0)
+			break;
+		c = *s++;
+	}
+
+	return tar_ret_errno(0, n);
+}
+
+static uintmax_t
+tar_atol(const char *s, size_t size, intmax_t min, uintmax_t max)
+{
+	const unsigned char *a = (const unsigned char *)s;
+
+	/* Check if it is a long two-complement base-256 number, positive or
+	 * negative. */
+	if (*a == 0xff || *a == 0x80)
+		return tar_atol256(s, size, min, max);
+	else
+		return tar_atol8(s, size);
+}
+
+uintmax_t
+tar_atoul(const char *s, size_t size, uintmax_t max)
+{
+	uintmax_t n = tar_atol(s, size, 0, UINTMAX_MAX);
+
+	if (n > max)
+		return tar_ret_errno(ERANGE, UINTMAX_MAX);
+
+	return n;
+}
+
+intmax_t
+tar_atosl(const char *s, size_t size, intmax_t min, intmax_t max)
+{
+	intmax_t n = tar_atol(s, size, INTMAX_MIN, INTMAX_MAX);
+
+	if (n < min)
+		return tar_ret_errno(ERANGE, INTMAX_MIN);
+	if (n > max)
+		return tar_ret_errno(ERANGE, INTMAX_MAX);
 
 	return n;
 }
 
 static char *
-get_prefix_name(struct tar_header *h)
+tar_header_get_prefix_name(struct tar_header *h)
 {
 	return str_fmt("%.*s/%.*s", (int)sizeof(h->prefix), h->prefix,
 	               (int)sizeof(h->name), h->name);
 }
 
 static mode_t
-get_unix_mode(struct tar_header *h)
+tar_header_get_unix_mode(struct tar_header *h)
 {
 	mode_t mode;
 	enum tar_filetype type;
@@ -122,7 +245,7 @@ get_unix_mode(struct tar_header *h)
 		break;
 	}
 
-	mode |= OtoM(h->mode, sizeof(h->mode));
+	mode |= TAR_ATOUL(h->mode, mode_t);
 
 	return mode;
 }
@@ -151,10 +274,12 @@ tar_header_checksum(struct tar_header *h)
 }
 
 static int
-tar_header_decode(struct tar_header *h, struct tar_entry *d)
+tar_header_decode(struct tar_header *h, struct tar_entry *d, struct dpkg_error *err)
 {
 	long checksum;
 
+	errno = 0;
+
 	if (memcmp(h->magic, TAR_MAGIC_GNU, 6) == 0)
 		d->format = TAR_FORMAT_GNU;
 	else if (memcmp(h->magic, TAR_MAGIC_USTAR, 6) == 0)
@@ -168,31 +293,50 @@ tar_header_decode(struct tar_header *h, struct tar_entry *d)
 
 	/* Concatenate prefix and name to support ustar style long names. */
 	if (d->format == TAR_FORMAT_USTAR && h->prefix[0] != '\0')
-		d->name = get_prefix_name(h);
+		d->name = tar_header_get_prefix_name(h);
 	else
 		d->name = m_strndup(h->name, sizeof(h->name));
 	d->linkname = m_strndup(h->linkname, sizeof(h->linkname));
-	d->stat.mode = get_unix_mode(h);
-	d->size = (off_t)OtoM(h->size, sizeof(h->size));
-	d->mtime = (time_t)OtoM(h->mtime, sizeof(h->mtime));
-	d->dev = makedev(OtoM(h->devmajor, sizeof(h->devmajor)),
-			 OtoM(h->devminor, sizeof(h->devminor)));
+	d->stat.mode = tar_header_get_unix_mode(h);
+	/* Even though off_t is signed, we use an unsigned parser here because
+	 * negative offsets are not allowed. */
+	d->size = TAR_ATOUL(h->size, off_t);
+	if (errno)
+		return dpkg_put_errno(err, _("invalid tar header size field"));
+	d->mtime = TAR_ATOSL(h->mtime, time_t);
+	if (errno)
+		return dpkg_put_errno(err, _("invalid tar header mtime field"));
+
+	if (d->type == TAR_FILETYPE_CHARDEV || d->type == TAR_FILETYPE_BLOCKDEV)
+		d->dev = makedev(TAR_ATOUL(h->devmajor, dev_t),
+		                 TAR_ATOUL(h->devminor, dev_t));
+	else
+		d->dev = makedev(0, 0);
 
 	if (*h->user)
 		d->stat.uname = m_strndup(h->user, sizeof(h->user));
 	else
 		d->stat.uname = NULL;
-	d->stat.uid = (uid_t)OtoM(h->uid, sizeof(h->uid));
+	d->stat.uid = TAR_ATOUL(h->uid, uid_t);
+	if (errno)
+		return dpkg_put_errno(err, _("invalid tar header uid field"));
 
 	if (*h->group)
 		d->stat.gname = m_strndup(h->group, sizeof(h->group));
 	else
 		d->stat.gname = NULL;
-	d->stat.gid = (gid_t)OtoM(h->gid, sizeof(h->gid));
+	d->stat.gid = TAR_ATOUL(h->gid, gid_t);
+	if (errno)
+		return dpkg_put_errno(err, _("invalid tar header gid field"));
 
-	checksum = OtoM(h->checksum, sizeof(h->checksum));
+	checksum = tar_atol8(h->checksum, sizeof(h->checksum));
+	if (errno)
+		return dpkg_put_errno(err, _("invalid tar header checksum field"));
 
-	return tar_header_checksum(h) == checksum;
+	if (tar_header_checksum(h) != checksum)
+		return dpkg_put_error(err, _("invalid tar header checksum"));
+
+	return 0;
 }
 
 /**
@@ -207,8 +351,7 @@ tar_header_decode(struct tar_header *h, struct tar_entry *d)
  *   bogus name or link.
  */
 static int
-tar_gnu_long(void *ctx, const struct tar_operations *ops, struct tar_entry *te,
-             char **longp)
+tar_gnu_long(struct tar_archive *tar, struct tar_entry *te, char **longp)
 {
 	char buf[TARBLKSZ];
 	char *bp;
@@ -221,14 +364,15 @@ tar_gnu_long(void *ctx, const struct tar_operations *ops, struct tar_entry *te,
 	for (long_read = te->size; long_read > 0; long_read -= TARBLKSZ) {
 		int copysize;
 
-		status = ops->read(ctx, buf, TARBLKSZ);
+		status = tar->ops->read(tar, buf, TARBLKSZ);
 		if (status == TARBLKSZ)
 			status = 0;
 		else {
 			/* Read partial header record? */
 			if (status > 0) {
 				errno = 0;
-				status = -1;
+				status = dpkg_put_error(&tar->err,
+				                        _("partially read tar header"));
 			}
 
 			/* If we didn't get TARBLKSZ bytes read, punt. */
@@ -264,6 +408,8 @@ tar_entry_destroy(struct tar_entry *te)
 	free(te->linkname);
 	free(te->stat.uname);
 	free(te->stat.gname);
+
+	memset(te, 0, sizeof(*te));
 }
 
 struct tar_symlink_entry {
@@ -295,7 +441,7 @@ tar_entry_update_from_system(struct tar_entry *te)
 }
 
 int
-tar_extractor(void *ctx, const struct tar_operations *ops)
+tar_extractor(struct tar_archive *tar)
 {
 	int status;
 	char buffer[TARBLKSZ];
@@ -313,17 +459,18 @@ tar_extractor(void *ctx, const struct tar_operations *ops)
 	h.stat.uname = NULL;
 	h.stat.gname = NULL;
 
-	while ((status = ops->read(ctx, buffer, TARBLKSZ)) == TARBLKSZ) {
+	while ((status = tar->ops->read(tar, buffer, TARBLKSZ)) == TARBLKSZ) {
 		int name_len;
 
-		if (!tar_header_decode((struct tar_header *)buffer, &h)) {
+		if (tar_header_decode((struct tar_header *)buffer, &h, &tar->err) < 0) {
 			if (h.name[0] == '\0') {
-				/* End of tape. */
+				/* The checksum failed on the terminating
+				 * End Of Tape block entry of zeros. */
+				dpkg_error_destroy(&tar->err);
+
+				/* End Of Tape. */
 				status = 0;
 			} else {
-				/* Indicates broken tarfile:
-				 * “Header checksum error”. */
-				errno = 0;
 				status = -1;
 			}
 			tar_entry_destroy(&h);
@@ -331,20 +478,24 @@ tar_extractor(void *ctx, const struct tar_operations *ops)
 		}
 		if (h.type != TAR_FILETYPE_GNU_LONGLINK &&
 		    h.type != TAR_FILETYPE_GNU_LONGNAME) {
-			if (next_long_name)
+			if (next_long_name) {
+				free(h.name);
 				h.name = next_long_name;
+			}
 
-			if (next_long_link)
+			if (next_long_link) {
+				free(h.linkname);
 				h.linkname = next_long_link;
+			}
 
 			next_long_link = NULL;
 			next_long_name = NULL;
 		}
 
 		if (h.name[0] == '\0') {
-			/* Indicates broken tarfile: “Bad header data”. */
+			status = dpkg_put_error(&tar->err,
+			                        _("invalid tar header with empty name field"));
 			errno = 0;
-			status = -1;
 			tar_entry_destroy(&h);
 			break;
 		}
@@ -355,7 +506,7 @@ tar_extractor(void *ctx, const struct tar_operations *ops)
 		case TAR_FILETYPE_FILE:
 			/* Compatibility with pre-ANSI ustar. */
 			if (h.name[name_len - 1] != '/') {
-				status = ops->extract_file(ctx, &h);
+				status = tar->ops->extract_file(tar, &h);
 				break;
 			}
 			/* Else, fall through. */
@@ -363,10 +514,10 @@ tar_extractor(void *ctx, const struct tar_operations *ops)
 			if (h.name[name_len - 1] == '/') {
 				h.name[name_len - 1] = '\0';
 			}
-			status = ops->mkdir(ctx, &h);
+			status = tar->ops->mkdir(tar, &h);
 			break;
 		case TAR_FILETYPE_HARDLINK:
-			status = ops->link(ctx, &h);
+			status = tar->ops->link(tar, &h);
 			break;
 		case TAR_FILETYPE_SYMLINK:
 			symlink_node = m_malloc(sizeof(*symlink_node));
@@ -383,18 +534,42 @@ tar_extractor(void *ctx, const struct tar_operations *ops)
 		case TAR_FILETYPE_CHARDEV:
 		case TAR_FILETYPE_BLOCKDEV:
 		case TAR_FILETYPE_FIFO:
-			status = ops->mknod(ctx, &h);
+			status = tar->ops->mknod(tar, &h);
 			break;
 		case TAR_FILETYPE_GNU_LONGLINK:
-			status = tar_gnu_long(ctx, ops, &h, &next_long_link);
+			status = tar_gnu_long(tar, &h, &next_long_link);
 			break;
 		case TAR_FILETYPE_GNU_LONGNAME:
-			status = tar_gnu_long(ctx, ops, &h, &next_long_name);
+			status = tar_gnu_long(tar, &h, &next_long_name);
+			break;
+		case TAR_FILETYPE_GNU_VOLUME:
+		case TAR_FILETYPE_GNU_MULTIVOL:
+		case TAR_FILETYPE_GNU_SPARSE:
+		case TAR_FILETYPE_GNU_DUMPDIR:
+			status = dpkg_put_error(&tar->err,
+			                        _("unsupported GNU tar header type '%c'"),
+			                        h.type);
+			errno = 0;
+			break;
+		case TAR_FILETYPE_SOLARIS_EXTENDED:
+		case TAR_FILETYPE_SOLARIS_ACL:
+			status = dpkg_put_error(&tar->err,
+			                        _("unsupported Solaris tar header type '%c'"),
+			                        h.type);
+			errno = 0;
+			break;
+		case TAR_FILETYPE_PAX_GLOBAL:
+		case TAR_FILETYPE_PAX_EXTENDED:
+			status = dpkg_put_error(&tar->err,
+			                        _("unsupported PAX tar header type '%c'"),
+			                        h.type);
+			errno = 0;
 			break;
 		default:
-			/* Indicates broken tarfile: “Bad header field”. */
+			status = dpkg_put_error(&tar->err,
+			                        _("unknown tar header type '%c'"),
+			                        h.type);
 			errno = 0;
-			status = -1;
 		}
 		tar_entry_destroy(&h);
 		if (status != 0)
@@ -405,7 +580,7 @@ tar_extractor(void *ctx, const struct tar_operations *ops)
 	while (symlink_head) {
 		symlink_node = symlink_head->next;
 		if (status == 0)
-			status = ops->symlink(ctx, &symlink_head->h);
+			status = tar->ops->symlink(tar, &symlink_head->h);
 		tar_entry_destroy(&symlink_head->h);
 		free(symlink_head);
 		symlink_head = symlink_node;
@@ -416,11 +591,11 @@ tar_extractor(void *ctx, const struct tar_operations *ops)
 	free(next_long_link);
 
 	if (status > 0) {
-		/* Indicates broken tarfile: “Read partial header record”. */
+		status = dpkg_put_error(&tar->err,
+		                        _("partially read tar header"));
 		errno = 0;
-		return -1;
-	} else {
-		/* Return whatever I/O function returned. */
-		return status;
 	}
+
+	/* Return whatever I/O function returned. */
+	return status;
 }
