@@ -2,8 +2,8 @@
  * dpkg - main program for package management
  * enquiry.c - status enquiry and listing options
  *
- * Copyright © 1995,1996 Ian Jackson <ian@chiark.greenend.org.uk>
- * Copyright © 2006, 2008-2015 Guillem Jover <guillem@debian.org>
+ * Copyright © 1995,1996 Ian Jackson <ijackson@chiark.greenend.org.uk>
+ * Copyright © 2006, 2008-2016 Guillem Jover <guillem@debian.org>
  * Copyright © 2011 Linaro Limited
  * Copyright © 2011 Raphaël Hertzog <hertzog@debian.org>
  *
@@ -26,7 +26,6 @@
 
 #include <sys/types.h>
 
-#include <assert.h>
 #include <string.h>
 #include <fcntl.h>
 #include <unistd.h>
@@ -40,11 +39,12 @@
 #include <dpkg/arch.h>
 #include <dpkg/pkg-array.h>
 #include <dpkg/pkg-show.h>
+#include <dpkg/triglib.h>
 #include <dpkg/string.h>
 #include <dpkg/options.h>
+#include <dpkg/db-ctrl.h>
+#include <dpkg/db-fsys.h>
 
-#include "filesdb.h"
-#include "infodb.h"
 #include "main.h"
 
 struct audit_problem {
@@ -169,7 +169,7 @@ static void describebriefly(struct pkginfo *pkg) {
   l= strlen(pkg->set->name);
   if (l>20) maxl -= (l-20);
 
-  pdesc = pkg_summary(pkg, &pkg->installed, &l);
+  pdesc = pkgbin_synopsis(pkg, &pkg->installed, &l);
   l = min(l, maxl);
 
   printf(" %-20s %.*s\n", pkg_name(pkg, pnaw_nonambig), l, pdesc);
@@ -198,7 +198,7 @@ audit(const char *const *argv)
   modstatdb_open(msdbrw_readonly);
 
   if (!*argv)
-    pkg_array_init_from_db(&array);
+    pkg_array_init_from_hash(&array);
   else
     pkg_array_init_from_names(&array, pkg_array_mapper, (const char **)argv);
 
@@ -273,7 +273,7 @@ unpackchk(const char *const *argv)
 {
   int totalcount, sects;
   struct sectionentry *sectionentries, *se, **sep;
-  struct pkgiterator *it;
+  struct pkg_hash_iter *iter;
   struct pkginfo *pkg;
   const char *thissect;
   char buf[20];
@@ -287,12 +287,12 @@ unpackchk(const char *const *argv)
   totalcount= 0;
   sectionentries = NULL;
   sects= 0;
-  it = pkg_db_iter_new();
-  while ((pkg = pkg_db_iter_next_pkg(it))) {
+  iter = pkg_hash_iter_new();
+  while ((pkg = pkg_hash_iter_next_pkg(iter))) {
     if (!yettobeunpacked(pkg, &thissect)) continue;
     for (se= sectionentries; se && strcasecmp(thissect,se->name); se= se->next);
     if (!se) {
-      se= nfmalloc(sizeof(struct sectionentry));
+      se = nfmalloc(sizeof(*se));
       for (sep= &sectionentries;
            *sep && strcasecmp(thissect,(*sep)->name) > 0;
            sep= &(*sep)->next);
@@ -304,27 +304,27 @@ unpackchk(const char *const *argv)
     }
     se->count++; totalcount++;
   }
-  pkg_db_iter_free(it);
+  pkg_hash_iter_free(iter);
 
   if (totalcount == 0)
     return 0;
 
   if (totalcount <= 12) {
-    it = pkg_db_iter_new();
-    while ((pkg = pkg_db_iter_next_pkg(it))) {
+    iter = pkg_hash_iter_new();
+    while ((pkg = pkg_hash_iter_next_pkg(iter))) {
       if (!yettobeunpacked(pkg, NULL))
         continue;
       describebriefly(pkg);
     }
-    pkg_db_iter_free(it);
+    pkg_hash_iter_free(iter);
   } else if (sects <= 12) {
     for (se= sectionentries; se; se= se->next) {
       sprintf(buf,"%d",se->count);
       printf(_(" %d in %s: "),se->count,se->name);
       width= 70-strlen(se->name)-strlen(buf);
       while (width > 59) { putchar(' '); width--; }
-      it = pkg_db_iter_new();
-      while ((pkg = pkg_db_iter_next_pkg(it))) {
+      iter = pkg_hash_iter_new();
+      while ((pkg = pkg_hash_iter_next_pkg(iter))) {
         const char *pkgname;
 
         if (!yettobeunpacked(pkg,&thissect)) continue;
@@ -335,7 +335,7 @@ unpackchk(const char *const *argv)
         if (width < 4) { printf(" ..."); break; }
         printf(" %s", pkgname);
       }
-      pkg_db_iter_free(it);
+      pkg_hash_iter_free(iter);
       putchar('\n');
     }
   } else {
@@ -357,87 +357,113 @@ unpackchk(const char *const *argv)
   return 0;
 }
 
+static const struct assert_feature {
+  const char *name;
+  const char *desc;
+  const char *version;
+} assert_features[] = {
+  {
+    .name = "support-predepends",
+    .desc = N_("the Pre-Depends field"),
+    .version = "1.1.0",
+  }, {
+    .name = "working-epoch",
+    .desc = N_("epochs in versions"),
+    .version = "1.4.0.7",
+  }, {
+    .name = "long-filenames",
+    .desc = N_("long filenames in .deb archives"),
+    .version = "1.4.1.17",
+  }, {
+    .name = "multi-conrep",
+    .desc = N_("multiple Conflicts and Replaces"),
+    .version = "1.4.1.19",
+  }, {
+    .name = "multi-arch",
+    .desc = N_("multi-arch fields and semantics"),
+    .version = "1.16.2",
+  }, {
+    .name = "versioned-provides",
+    .desc = N_("versioned relationships in the Provides field"),
+    .version = "1.17.11",
+  }, {
+    .name = "protected-field",
+    .desc = N_("the Protected field"),
+    .version = "1.20.1",
+  }, {
+    .name = NULL,
+  }
+};
+
 static int
 assert_version_support(const char *const *argv,
-                       struct dpkg_version *version,
-                       const char *feature_name)
+                       const struct assert_feature *feature)
 {
-  struct pkginfo *pkg;
+  const char *running_version_str;
+  struct dpkg_version running_version = DPKG_VERSION_INIT;
+  struct dpkg_version version = { 0, feature->version, NULL };
+  struct dpkg_error err = DPKG_ERROR_INIT;
 
   if (*argv)
     badusage(_("--%s takes no arguments"), cipaction->olong);
 
-  modstatdb_open(msdbrw_readonly);
+  /*
+   * When using the feature asserts, we want to know whether the currently
+   * running dpkg, which we might be running under (say from within a
+   * maintainer script) and which might have a different version, supports
+   * the requested feature. As dpkg is an Essential package, it is expected
+   * to work even when just unpacked, and so its own version is enough.
+   */
+  running_version_str = getenv("DPKG_RUNNING_VERSION");
 
-  pkg = pkg_db_find_singleton("dpkg");
-  switch (pkg->status) {
-  case PKG_STAT_INSTALLED:
-  case PKG_STAT_TRIGGERSPENDING:
+  /*
+   * If we are not running from within a maintainer script, then that means
+   * once we do, the executed dpkg will support the requested feature, if
+   * we know about it. Always return success in that case.
+   */
+  if (str_is_unset(running_version_str))
     return 0;
-  case PKG_STAT_UNPACKED:
-  case PKG_STAT_HALFCONFIGURED:
-  case PKG_STAT_HALFINSTALLED:
-  case PKG_STAT_TRIGGERSAWAITED:
-    if (dpkg_version_relate(&pkg->configversion, DPKG_RELATION_GE, version))
-      return 0;
-    printf(_("Version of dpkg with working %s support not yet configured.\n"
-             " Please use 'dpkg --configure dpkg', and then try again.\n"),
-           feature_name);
-    return 1;
-  default:
-    printf(_("dpkg not recorded as installed, cannot check for %s support!\n"),
-           feature_name);
-    return 1;
+
+  if (parseversion(&running_version, running_version_str, &err) < 0)
+    ohshit(_("cannot parse dpkg running version '%s': %s"),
+           running_version_str, err.str);
+
+  if (dpkg_version_relate(&running_version, DPKG_RELATION_GE, &version))
+    return 0;
+
+  printf(_("Running version of dpkg does not support %s.\n"
+           " Please upgrade to at least dpkg %s, and then try again.\n"),
+         feature->desc, versiondescribe(&version, vdew_nonambig));
+  return 1;
+}
+
+const char *assert_feature_name;
+
+int
+assert_feature(const char *const *argv)
+{
+  const struct assert_feature *feature;
+
+  if (strcmp(assert_feature_name, "help") == 0) {
+    printf(_("%s assert options - assert whether features are supported:\n"),
+           dpkg_get_progname());
+
+    for (feature = assert_features; feature->name; feature++) {
+      printf("  %-19s %-9s %s\n", feature->name, feature->version,
+             gettext(feature->desc));
+    }
+
+    exit(0);
   }
-}
 
-int
-assertpredep(const char *const *argv)
-{
-  struct dpkg_version version = { 0, "1.1.0", NULL };
+  for (feature = assert_features; feature->name; feature++) {
+    if (strcmp(feature->name, assert_feature_name) != 0)
+      continue;
 
-  return assert_version_support(argv, &version, _("Pre-Depends field"));
-}
+    return assert_version_support(argv, feature);
+  }
 
-int
-assertepoch(const char *const *argv)
-{
-  struct dpkg_version version = { 0, "1.4.0.7", NULL };
-
-  return assert_version_support(argv, &version, _("epoch"));
-}
-
-int
-assertlongfilenames(const char *const *argv)
-{
-  struct dpkg_version version = { 0, "1.4.1.17", NULL };
-
-  return assert_version_support(argv, &version, _("long filenames"));
-}
-
-int
-assertmulticonrep(const char *const *argv)
-{
-  struct dpkg_version version = { 0, "1.4.1.19", NULL };
-
-  return assert_version_support(argv, &version,
-                                _("multiple Conflicts and Replaces"));
-}
-
-int
-assertmultiarch(const char *const *argv)
-{
-  struct dpkg_version version = { 0, "1.16.2", NULL };
-
-  return assert_version_support(argv, &version, _("multi-arch"));
-}
-
-int
-assertverprovides(const char *const *argv)
-{
-  struct dpkg_version version = { 0, "1.17.11", NULL };
-
-  return assert_version_support(argv, &version, _("versioned Provides"));
+  badusage(_("unknown --%s-<feature>"), cipaction->olong);
 }
 
 /**
@@ -458,7 +484,7 @@ predeppackage(const char *const *argv)
 {
   static struct varbuf vb;
 
-  struct pkgiterator *it;
+  struct pkg_hash_iter *iter;
   struct pkginfo *pkg = NULL, *startpkg, *trypkg;
   struct dependency *dep;
   struct deppossi *possi, *provider;
@@ -471,13 +497,13 @@ predeppackage(const char *const *argv)
   clear_istobes();
 
   dep = NULL;
-  it = pkg_db_iter_new();
-  while (!dep && (pkg = pkg_db_iter_next_pkg(it))) {
+  iter = pkg_hash_iter_new();
+  while (!dep && (pkg = pkg_hash_iter_next_pkg(iter))) {
     /* Ignore packages user doesn't want. */
     if (pkg->want != PKG_WANT_INSTALL)
       continue;
     /* Ignore packages not available. */
-    if (!pkg->files)
+    if (!pkg->archives)
       continue;
     pkg->clientdata->istobe = PKG_ISTOBE_PREINSTALL;
     for (dep= pkg->available.depends; dep; dep= dep->next) {
@@ -490,11 +516,13 @@ predeppackage(const char *const *argv)
     pkg->clientdata->istobe = PKG_ISTOBE_NORMAL;
     /* If dep is NULL we go and get the next package. */
   }
-  pkg_db_iter_free(it);
+  pkg_hash_iter_free(iter);
 
   if (!dep)
     return 1; /* Not found. */
-  assert(pkg);
+  if (pkg == NULL)
+    internerr("unexpected unfound package");
+
   startpkg= pkg;
   pkg->clientdata->istobe = PKG_ISTOBE_PREINSTALL;
 
@@ -510,7 +538,7 @@ predeppackage(const char *const *argv)
 
       possi_iter = deppossi_pkg_iter_new(possi, wpb_available);
       while (!pkg && (trypkg = deppossi_pkg_iter_next(possi_iter))) {
-        if (trypkg->files &&
+        if (trypkg->archives &&
             trypkg->clientdata->istobe == PKG_ISTOBE_NORMAL &&
             versionsatisfied(&trypkg->available, possi)) {
           pkg = trypkg;
@@ -524,7 +552,7 @@ predeppackage(const char *const *argv)
           if (!pkg_virtual_deppossi_satisfied(possi, provider))
             continue;
           trypkg = provider->up->up;
-          if (!trypkg->files)
+          if (!trypkg->archives)
             continue;
           if (trypkg->clientdata->istobe == PKG_ISTOBE_NORMAL) {
             pkg = trypkg;
@@ -574,14 +602,6 @@ printarch(const char *const *argv)
   return 0;
 }
 
-int
-printinstarch(const char *const *argv)
-{
-  warning(_("obsolete option '--%s'; please use '--%s' instead"),
-          "print-installation-architecture", "print-architecture");
-  return printarch(argv);
-}
-
 int
 print_foreign_arches(const char *const *argv)
 {
@@ -604,6 +624,70 @@ print_foreign_arches(const char *const *argv)
   return 0;
 }
 
+int
+validate_pkgname(const char *const *argv)
+{
+  const char *emsg;
+
+  if (!argv[0] || argv[1])
+    badusage(_("--%s takes one <pkgname> argument"), cipaction->olong);
+
+  emsg = pkg_name_is_illegal(argv[0]);
+  if (emsg)
+    ohshit(_("package name '%s' is invalid: %s"), argv[0], emsg);
+
+  return 0;
+}
+
+int
+validate_trigname(const char *const *argv)
+{
+  const char *emsg;
+
+  if (!argv[0] || argv[1])
+    badusage(_("--%s takes one <trigname> argument"), cipaction->olong);
+
+  emsg = trig_name_is_illegal(argv[0]);
+  if (emsg)
+    ohshit(_("trigger name '%s' is invalid: %s"), argv[0], emsg);
+
+  return 0;
+}
+
+int
+validate_archname(const char *const *argv)
+{
+  const char *emsg;
+
+  if (!argv[0] || argv[1])
+    badusage(_("--%s takes one <archname> argument"), cipaction->olong);
+
+  emsg = dpkg_arch_name_is_illegal(argv[0]);
+  if (emsg)
+    ohshit(_("architecture name '%s' is invalid: %s"), argv[0], emsg);
+
+  return 0;
+}
+
+int
+validate_version(const char *const *argv)
+{
+  struct dpkg_version version;
+  struct dpkg_error err;
+
+  if (!argv[0] || argv[1])
+    badusage(_("--%s takes one <version> argument"), cipaction->olong);
+
+  if (parseversion(&version, argv[0], &err) < 0) {
+    dpkg_error_print(&err, _("version '%s' has bad syntax"), argv[0]);
+    dpkg_error_destroy(&err);
+
+    return 1;
+  }
+
+  return 0;
+}
+
 int
 cmpversions(const char *const *argv)
 {
@@ -684,8 +768,8 @@ cmpversions(const char *const *argv)
   }
   rc = dpkg_version_compare(&a, &b);
   debug(dbg_general, "cmpversions a='%s' b='%s' r=%d",
-        versiondescribe(&a,vdew_always),
-        versiondescribe(&b,vdew_always),
+        versiondescribe_c(&a,vdew_always),
+        versiondescribe_c(&b,vdew_always),
         rc);
   if (rc > 0)
     return rip->if_greater;
