@@ -2,7 +2,7 @@
  * dpkg-statoverride - override ownership and mode of files
  *
  * Copyright © 2000, 2001 Wichert Akkerman <wakkerma@debian.org>
- * Copyright © 2006-2012 Guillem Jover <guillem@debian.org>
+ * Copyright © 2006-2015 Guillem Jover <guillem@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -40,13 +40,16 @@
 #include <dpkg/i18n.h>
 #include <dpkg/dpkg.h>
 #include <dpkg/dpkg-db.h>
+#include <dpkg/string.h>
 #include <dpkg/path.h>
 #include <dpkg/dir.h>
 #include <dpkg/glob.h>
+#include <dpkg/db-fsys.h>
 #include <dpkg/options.h>
 
-#include "main.h"
-#include "filesdb.h"
+#include "force.h"
+#include "actions.h"
+#include "security-mac.h"
 
 static const char printforhelp[] = N_(
 "Use --help for help about overriding file stat information.");
@@ -55,7 +58,7 @@ static void DPKG_ATTR_NORET
 printversion(const struct cmdinfo *cip, const char *value)
 {
 	printf(_("Debian %s version %s.\n"), dpkg_get_progname(),
-	       DPKG_VERSION_ARCH);
+	       PACKAGE_RELEASE);
 
 	printf(_(
 "This is free software; see the GNU General Public License version 2 or\n"
@@ -84,8 +87,13 @@ usage(const struct cmdinfo *cip, const char *value)
 	printf(_(
 "Options:\n"
 "  --admindir <directory>   set the directory with the statoverride file.\n"
+"  --instdir <directory>    set the root directory, but not the admin dir.\n"
+"  --root <directory>       set the directory of the root filesystem.\n"
 "  --update                 immediately update <path> permissions.\n"
-"  --force                  force an action even if a sanity check fails.\n"
+"  --force                  deprecated alias for --force-all.\n"
+"  --force-<thing>[,...]    override problems (see --force-help).\n"
+"  --no-force-<thing>[,...] stop when problems encountered.\n"
+"  --refuse-<thing>[,...]   ditto.\n"
 "  --quiet                  quiet operation, minimal output.\n"
 "  --help                   show this help message.\n"
 "  --version                show the version.\n"
@@ -96,12 +104,28 @@ usage(const struct cmdinfo *cip, const char *value)
 	exit(0);
 }
 
+#define FORCE_STATCMD_MASK \
+	FORCE_SECURITY_MAC | FORCE_STATOVERRIDE_ADD | FORCE_STATOVERRIDE_DEL
+
 static const char *admindir;
+const char *instdir;
 
 static int opt_verbose = 1;
-static int opt_force = 0;
 static int opt_update = 0;
 
+static void
+set_instdir(const struct cmdinfo *cip, const char *value)
+{
+	instdir = dpkg_fsys_set_dir(value);
+}
+
+static void
+set_root(const struct cmdinfo *cip, const char *value)
+{
+	instdir = dpkg_fsys_set_dir(value);
+	admindir = dpkg_fsys_get_path(ADMINDIR);
+}
+
 static char *
 path_cleanup(const char *path)
 {
@@ -122,7 +146,13 @@ statdb_node_new(const char *user, const char *group, const char *mode)
 	filestat = nfmalloc(sizeof(*filestat));
 
 	filestat->uid = statdb_parse_uid(user);
+	if (filestat->uid == (uid_t)-1)
+		ohshit(_("user '%s' does not exist"), user);
+	filestat->uname = NULL;
 	filestat->gid = statdb_parse_gid(group);
+	if (filestat->gid == (gid_t)-1)
+		ohshit(_("group '%s' does not exist"), group);
+	filestat->gname = NULL;
 	filestat->mode = statdb_parse_mode(mode);
 
 	return filestat;
@@ -131,9 +161,9 @@ statdb_node_new(const char *user, const char *group, const char *mode)
 static struct file_stat **
 statdb_node_find(const char *filename)
 {
-	struct filenamenode *file;
+	struct fsys_namenode *file;
 
-	file = findnamenode(filename, 0);
+	file = fsys_hash_find_node(filename, 0);
 
 	return &file->statoverride;
 }
@@ -141,10 +171,10 @@ statdb_node_find(const char *filename)
 static int
 statdb_node_remove(const char *filename)
 {
-	struct filenamenode *file;
+	struct fsys_namenode *file;
 
-	file = findnamenode(filename, fnn_nonew);
-	if (!file || (file && !file->statoverride))
+	file = fsys_hash_find_node(filename, FHFF_NONE);
+	if (!file || !file->statoverride)
 		return 0;
 
 	file->statoverride = NULL;
@@ -156,13 +186,17 @@ static void
 statdb_node_apply(const char *filename, struct file_stat *filestat)
 {
 	if (chown(filename, filestat->uid, filestat->gid) < 0)
-		ohshite(_("error setting ownership of `%.255s'"), filename);
-	if (chmod(filename, filestat->mode))
-		ohshite(_("error setting permissions of `%.255s'"), filename);
+		ohshite(_("error setting ownership of '%.255s'"), filename);
+	if (chmod(filename, filestat->mode & ~S_IFMT))
+		ohshite(_("error setting permissions of '%.255s'"), filename);
+
+	dpkg_selabel_load();
+	dpkg_selabel_set_context(filename, filename, filestat->mode);
+	dpkg_selabel_close();
 }
 
 static void
-statdb_node_print(FILE *out, struct filenamenode *file)
+statdb_node_print(FILE *out, struct fsys_namenode *file)
 {
 	struct file_stat *filestat = file->statoverride;
 	struct passwd *pw;
@@ -174,16 +208,20 @@ statdb_node_print(FILE *out, struct filenamenode *file)
 	pw = getpwuid(filestat->uid);
 	if (pw)
 		fprintf(out, "%s ", pw->pw_name);
+	else if (filestat->uname)
+		fprintf(out, "%s ", filestat->uname);
 	else
 		fprintf(out, "#%d ", filestat->uid);
 
 	gr = getgrgid(filestat->gid);
 	if (gr)
 		fprintf(out, "%s ", gr->gr_name);
+	else if (filestat->gname)
+		fprintf(out, "%s ", filestat->gname);
 	else
 		fprintf(out, "#%d ", filestat->gid);
 
-	fprintf(out, "%o %s\n", filestat->mode, file->name);
+	fprintf(out, "%o %s\n", filestat->mode & ~S_IFMT, file->name);
 }
 
 static void
@@ -191,17 +229,17 @@ statdb_write(void)
 {
 	char *dbname;
 	struct atomic_file *dbfile;
-	struct fileiterator *iter;
-	struct filenamenode *file;
+	struct fsys_hash_iter *iter;
+	struct fsys_namenode *file;
 
 	dbname = dpkg_db_get_path(STATOVERRIDEFILE);
-	dbfile = atomic_file_new(dbname, aff_backup);
+	dbfile = atomic_file_new(dbname, ATOMIC_FILE_BACKUP);
 	atomic_file_open(dbfile);
 
-	iter = files_db_iter_new();
-	while ((file = files_db_iter_next(iter)))
+	iter = fsys_hash_iter_new();
+	while ((file = fsys_hash_iter_next(iter)))
 		statdb_node_print(dbfile->fp, file);
-	files_db_iter_free(iter);
+	fsys_hash_iter_free(iter);
 
 	atomic_file_sync(dbfile);
 	atomic_file_close(dbfile);
@@ -233,7 +271,7 @@ statoverride_add(const char *const *argv)
 
 	filestat = statdb_node_find(filename);
 	if (*filestat != NULL) {
-		if (opt_force)
+		if (in_force(FORCE_STATOVERRIDE_ADD))
 			warning(_("an override for '%s' already exists, "
 			          "but --force specified so will be ignored"),
 			        filename);
@@ -246,12 +284,21 @@ statoverride_add(const char *const *argv)
 
 	if (opt_update) {
 		struct stat st;
+		struct varbuf realfilename = VARBUF_INIT;
 
-		if (stat(filename, &st) == 0)
-			statdb_node_apply(filename, *filestat);
-		else if (opt_verbose)
+		varbuf_add_str(&realfilename, instdir);
+		varbuf_add_str(&realfilename, filename);
+		varbuf_end_str(&realfilename);
+
+		if (stat(realfilename.buf, &st) == 0) {
+			(*filestat)->mode |= st.st_mode & S_IFMT;
+			statdb_node_apply(realfilename.buf, *filestat);
+		} else if (opt_verbose) {
 			warning(_("--update given but %s does not exist"),
-			        filename);
+			        realfilename.buf);
+		}
+
+		varbuf_destroy(&realfilename);
 	}
 
 	statdb_write();
@@ -275,7 +322,7 @@ statoverride_remove(const char *const *argv)
 	if (!statdb_node_remove(filename)) {
 		if (opt_verbose)
 			warning(_("no override present"));
-		if (opt_force)
+		if (in_force(FORCE_STATOVERRIDE_DEL))
 			return 0;
 		else
 			return 2;
@@ -294,8 +341,8 @@ statoverride_remove(const char *const *argv)
 static int
 statoverride_list(const char *const *argv)
 {
-	struct fileiterator *iter;
-	struct filenamenode *file;
+	struct fsys_hash_iter *iter;
+	struct fsys_namenode *file;
 	const char *thisarg;
 	struct glob_node *glob_list = NULL;
 	int ret = 1;
@@ -308,8 +355,8 @@ statoverride_list(const char *const *argv)
 	if (glob_list == NULL)
 		glob_list_prepend(&glob_list, m_strdup("*"));
 
-	iter = files_db_iter_new();
-	while ((file = files_db_iter_next(iter))) {
+	iter = fsys_hash_iter_new();
+	while ((file = fsys_hash_iter_next(iter))) {
 		struct glob_node *g;
 
 		for (g = glob_list; g; g = g->next) {
@@ -320,21 +367,34 @@ statoverride_list(const char *const *argv)
 			}
 		}
 	}
-	files_db_iter_free(iter);
+	fsys_hash_iter_free(iter);
 
 	glob_list_free(glob_list);
 
 	return ret;
 }
 
+static void
+set_force_obsolete(const struct cmdinfo *cip, const char *value)
+{
+	warning(_("deprecated --%s option; use --%s instead"),
+	        cip->olong, "force-all");
+	set_force(FORCE_ALL);
+}
+
 static const struct cmdinfo cmdinfos[] = {
 	ACTION("add",    0, act_install,   statoverride_add),
 	ACTION("remove", 0, act_remove,    statoverride_remove),
 	ACTION("list",   0, act_listfiles, statoverride_list),
 
 	{ "admindir",   0,   1,  NULL,         &admindir, NULL          },
+	{ "instdir",    0,   1,  NULL,         NULL,      set_instdir,  0 },
+	{ "root",       0,   1,  NULL,         NULL,      set_root,     0 },
 	{ "quiet",      0,   0,  &opt_verbose, NULL,      NULL, 0       },
-	{ "force",      0,   0,  &opt_force,   NULL,      NULL, 1       },
+	{ "force",      0,   0,  NULL,         NULL,      set_force_obsolete },
+	{ "force",      0,   2,  NULL,         NULL,      set_force_option, 1 },
+	{ "no-force",   0,   2,  NULL,         NULL,      set_force_option, 0 },
+	{ "refuse",     0,   2,  NULL,         NULL,      set_force_option, 0 },
 	{ "update",     0,   0,  &opt_update,  NULL,      NULL, 1       },
 	{ "help",       '?', 0,  NULL,         NULL,      usage         },
 	{ "version",    0,   0,  NULL,         NULL,      printversion  },
@@ -348,19 +408,21 @@ main(int argc, const char *const *argv)
 
 	dpkg_locales_init(PACKAGE);
 	dpkg_program_init("dpkg-statoverride");
+	set_force_default(FORCE_STATCMD_MASK);
 	dpkg_options_parse(&argv, cmdinfos, printforhelp);
 
+	instdir = dpkg_fsys_set_dir(instdir);
 	admindir = dpkg_db_set_dir(admindir);
 
 	if (!cipaction)
 		badusage(_("need an action option"));
 
-	filesdbinit();
-	ensure_statoverrides();
+	ensure_statoverrides(STATDB_PARSE_LAX);
 
 	ret = cipaction->action(argv);
 
 	dpkg_program_done();
+	dpkg_locales_done();
 
 	return ret;
 }
